var Message = require('pomelo-protocol').Message;
var Constants = require('../../util/constants');
var logger = require('pomelo-logger').getLogger('pomelo', __filename);
var fs = require('fs');

var encode = function(reqId, route, msg, splitHeader) {
  if(!!reqId) {
    return composeResponse(this, reqId, route, msg, splitHeader);
  } else {
    return composePush(this, route, msg);
  }
};

var decode = function(msg) {
  msg = Message.decode(msg.body);
  var route = msg.route;

  // decode use dictionary
  if(!!msg.compressRoute) {
    if(!!this.connector.useDict) {
      var abbrs = this.dictionary.getAbbrs();
      if(!abbrs[route]) {
        logger.error('dictionary error! no abbrs for route : %s', route);
        return null;
      }
      route = msg.route = abbrs[route];
    } else {
      logger.error('fail to uncompress route code for msg: %j, server not enable dictionary.', msg);
      return null;
    }
  }

  // decode use protobuf
  if(!!this.protobuf && !!this.protobuf.getProtos().client[route]) {
    msg.body = this.protobuf.decode(route, msg.body);
  } else if(!!this.decodeIO_protobuf && !!this.decodeIO_protobuf.check(Constants.RESERVED.CLIENT, route)) {
    msg.body = this.decodeIO_protobuf.decode(route, msg.body);
  } else {
    try {
      msg.body = JSON.parse(msg.body.toString('utf8'));
    } catch (ex) {
      msg.body = {};
    }
  }

  return msg;
};

var composeResponse = function(server, msgId, route, msgBody, splitHeader) {
  if(!msgId || !route || !msgBody) {
    return null;
  }
  msgBody = encodeBody(server, route, msgBody, splitHeader);
  return Message.encode(msgId, Message.TYPE_RESPONSE, 0, null, msgBody);
};

var composePush = function(server, route, msgBody) {
  if(!route || !msgBody){
    return null;
  }
  msgBody = encodeBody(server, route, msgBody);
  // encode use dictionary
  var compressRoute = 0;
  if(!!server.dictionary) {
    var dict = server.dictionary.getDict();
    if(!!server.connector.useDict && !!dict[route]) {
      route = dict[route];
      compressRoute = 1;
    }
  }
  return Message.encode(0, Message.TYPE_PUSH, compressRoute, route, msgBody);
};

var isBasicType = function (type) {
    if (!type || type == 'object' || type == 'array') {
        return false;
    }
    return true;
}

var typeofValue = function (value) {
    var vt = typeof value;
    switch (vt) {
        case 'number':
            if (Number.isInteger(value)) {
                return 'int32';
            } else {
                return 'float';
            }
            break;
        case 'boolean':
            return 'int32';
            break;
        case 'string':
            if (isNaN(parseInt(value))) {
                return 'string';
            } else {
                return 'int32';
            }
            break;
        case 'undefined':
            break;
        case 'object':
            if (Object.prototype.toString.call(value) === '[object Array]') {
                return 'array';
            } else {
                return 'object';
            }
            break;
        case 'symbol':
            break;
        case 'function':
            break;
    }
    return null;
}

var getTypePriority  = function (fieldType)
{
  switch (fieldType) {
    case 'object':
      return 1;
    case 'array':
      return 2;
    case 'string':
      return 3;
    case 'float':
      return 4;
    case 'int32':
      return 5;
    default:
      return 6;

  }
}

var genProtobuf = function (myObject, prodef)
{
  prodef.__tags = prodef.__tags || {};
  prodef.__messages = prodef.__messages || {};
  var id = 0;
  var tagl = Object.keys(prodef.__tags).length;
  if (tagl > 0) {
      id = tagl;
  }
  for (var name in myObject) {
    if (!myObject.hasOwnProperty(name)) {
      continue;
    }
    var value = myObject[name];

      if (value === null) {
      continue;
    }
    if (value === undefined) {
        continue;
    }
    var vt = typeofValue(value);
    if (!vt) {
      continue;
    } else {
      var fieldDef = {};
      fieldDef.option = 'optional';
      id++;
      fieldDef.tag = id;
      fieldDef.name = name.toString();
      if (vt == 'object') {
        if (!prodef.__messages[name]) {
          prodef.__messages[name] = {};
        }
        fieldDef.type = name;
        genProtobuf(value, prodef.__messages[name])


      } else if (vt == 'array') {
        fieldDef.option = 'repeated';
          var isBaseicArray = true;
          var etype = null;
          for (var i = 0; i < value.length; i++) {
              var typarele = typeofValue(value[i]);
              if (!isBasicType(typarele)) {
                  isBaseicArray = false;
              } else {
                var tag1 = getTypePriority(etype);
                var tag2 = getTypePriority(typarele);

                if (tag1 > tag2) {
                  //类型升级
                  etype = typarele;
                }
                  // if (etype && etype != typarele) {
                  //     if (etype == 'string' || typarele == 'string') {
                  //         etype = 'string';
                  //     } else if (etype == 'float' || typarele == 'float') {
                  //         etype = 'float';
                  //     }
                  // } else {
                  //     etype = typarele;
                  // }
              }
          }
          if (isBaseicArray && etype) {
            fieldDef.type = etype;
          } else {
            if (!prodef.__messages[name]) {
              prodef.__messages[name] = {};
            }
            fieldDef.type = name;

            for (var i = 0; i < value.length; i++) {
                genProtobuf(value[i], prodef.__messages[name]);
            }
          }
      } else {
        fieldDef.type = vt;
      }
      if (prodef[name] == null) {
        prodef.__tags[id] = name;
        prodef[name] = fieldDef;
      } else {
        //升级协议
        var deforg = prodef[name];
        if (deforg.type != fieldDef.type) {
          //类型不同
          var tag1 = getTypePriority(deforg.type);
          var tag2 = getTypePriority(fieldDef.type);

          if (tag1 > tag2) {
            //类型升级
            deforg.type = fieldDef.type;
          }
        }
      }
    }
  }
}

var protoToJson = function (defs, json) {
    for (var objname in defs.__messages) {
      var cj = {};
      json['message ' + objname]  = cj;
      protoToJson(defs.__messages[objname], cj);
    }
    for (var objname in defs) {
      if (!defs.hasOwnProperty(objname)) {
        continue;
      }
      if (objname == '__messages') {
        continue;
      }
      if (objname == '__tags') {
        continue;
      }
      var fieldDef = defs[objname];
      var fieldname = fieldDef.option + ' ' + fieldDef.type + ' ' + fieldDef.name;
      json[fieldname] = fieldDef.tag;
    }
}

var encodeBody = function(server, route, msgBody, splitHeader) {
    // encode use protobuf
  if (!!server.protobuf && !!server.autoGenProtobuf && !server.protobuf.getProtos().server[route]) {
    var def = server.protobuf.autoProtos[route];
    if (!def) {
      server.protobuf.autoProtos[route] = {};
      def = server.protobuf.autoProtos[route];
    }
    genProtobuf(msgBody, def);
    var defJson = server.protobuf.autoProtosJson[route];
    if (!defJson) {
        server.protobuf.autoProtosJson[route] = {};
        defJson = server.protobuf.autoProtosJson[route];
    }
      if(!!def){
        protoToJson(def, defJson);


        var basePath = process.cwd() + '/autoProto';
        if (!fs.existsSync(basePath)) {
            fs.mkdirSync(basePath);
        }
        var autop = basePath + '/autoGenSvrProtos.json';
        fs.writeFileSync(autop, JSON.stringify(server.protobuf.autoProtosJson, null, 2), {flags: 'w'});

        try {
          var output = server.protobuf.encodeEx(def, msgBody);
          if(output && output.length > 0){
            var defo = new Buffer(JSON.stringify(def), 'utf8');
            logger.debug('test protobuf '+route + ' len ' + output.length + ' def len ' + defo.length);

          }
        } catch (e) {
          logger.error('test protobuf '+route);

          var autop1 = basePath + '/late_err_msg.json';
          fs.writeFileSync(autop1, JSON.stringify(msgBody, null, 2), {flags: 'w'});
        }

      }
  }
  var header = null;
  if (splitHeader) {
    header = splitHeader(msgBody, server.protobuf.encode);
    if (!header) {
      console.error('encode header fail');
    }
  }
  if (splitHeader && Object.keys(msgBody).length === 0) {
    msgBody = [];
  } else {
    if(!!server.protobuf && !!server.protobuf.getProtos().server[route]) {
      msgBody = server.protobuf.encode(route, msgBody);
    } else if(!!server.decodeIO_protobuf && !!server.decodeIO_protobuf.check(Constants.RESERVED.SERVER, route)) {
      msgBody = server.decodeIO_protobuf.encode(route, msgBody);
    } else {
      msgBody = new Buffer(JSON.stringify(msgBody), 'utf8');
    }
  }

  if (splitHeader) {
    var len = 4+header.length + (msgBody ? msgBody.length : 0);
    var buffer = new Buffer(len);
    buffer.writeUInt32LE(header.length);
    for (var i = 0; i < header.length;i++) {
      buffer[i + 4] = header[i];
    }
    if (msgBody) {
      for (var i = 0; i < msgBody.length;i++) {
        buffer[i + 4 + header.length] = msgBody[i];
      }
    }
    msgBody = buffer;
  }

  return msgBody;
};

module.exports = {
  encode: encode,
  decode: decode
};